
#include "datebang.h"

/* take these out if you're linking with CRT */ 
/*
#define strncpy lstrcpyn
inline char *strchr(const char *s,int c)
{ while (*s) if (*s==c) return const_cast<char *>(s); else ++s; return NULL; }
*/

/*

  AtVars

  Variables are of the format:
   ^@...^        regular
   ^@...\abc^    escape chars abc and \
   ^@...\\^      escape chars ^ and \

  The ... bits are interpreted by the LookupVar callback, which
  must stop reading at a VARCHAR or ESCCHAR, and must not write
  maxchars to dest.  LookupVar should return the number of chars
  written to dest, which does not need to be NULL terminated.
  If a sequence is encountered with LookupVar does not understand,
  it should return 0.

*/


#define VARCHAR '^'
#define ATCHAR '@'
#define ESCCHAR '\\'


/* maxchars includes the nCh that we are processing */ 
int EscapeVar(char *dest,int maxchars,int nCh,const char *escape)
{
  char chrs[256];
  char *ptr;
  int nChOut = 0;

  ASSERT(READABLE(dest,maxchars*sizeof(char)));
  ASSERT(WRITEABLE(dest,maxchars*sizeof(char)));
  ASSERT(READABLE(escape,nCh*sizeof(char)));

  /* get the list of chars to escape */ 
  ptr = chrs;
  /* always escape ESCCHAR */ 
  *ptr++ = ESCCHAR;
  while (*escape!=VARCHAR)
  {
    /* since ESCCHAR is always escaped */ 
    /* and VARCHAR can't be specified, */ 
    /* use ESCCHAR to escape VARCHAR   */ 
    if (*escape==ESCCHAR)
      *ptr++ = VARCHAR;
    else
      *ptr++ = *escape;

    ++escape;
  }
  *ptr = 0;

  /* copy original */ 
  ptr = new char[nCh+1];
  strncpy(ptr,dest,nCh);
  ptr[nCh] = 0; /* strncpy won't terminate it */ 
  /* could use strncpy(,,nCh+1); but this is safer */ 

  /* copy back to dest with escapes */ 
  nCh = 0;
  while (*ptr && nChOut<(maxchars-1))
  {
    /* escape this char? */ 
    if (strchr(chrs,*ptr))
    {
      *dest++ = ESCCHAR;
      ++nChOut;
    }
    if (nChOut==(maxchars-1))
      { --nChOut; break; }
    *dest++ = *ptr++;
    ++nChOut; ++nCh;
  }
  /* restore ptr */ 
  ptr -= nCh;
  delete[] ptr;

  return nChOut;
}

inline bool IsKeyDown(int key) { return (GetKeyState(key)&0x8000)!=0; }

int LookupStdVar(char *dest,int maxchars,LPCSTR src)
{
  bool left,right;
  int i;

  ASSERT(maxchars>1);
  ASSERT(WRITEABLE(dest,maxchars*sizeof(char)));
  ASSERT(READABLE(src,sizeof(char)));

  switch (*src++)
  {
  case '#': /* special char */ 
    switch (*src)
    {
    case 'A':
    case 'a': *dest = '\''; break;
    case 'C':
    case 'c': *dest = '^'; break;
    case 'D':
    case 'd': *dest = '$'; break;
    case 'N':
    case 'n': *dest = '\n'; break;
    case 'Q':
    case 'q': *dest = '"'; break;
    default:  *dest = *src;
    }
    return 1;

  case 'K':
	case 'k': /* keys */ 
		switch (*src++) {
    case 'A':
		case 'a': /* alt              */ 
			left = IsKeyDown(VK_LMENU);
			right = IsKeyDown(VK_RMENU);
			break;
    case 'C':
		case 'c': /* ctrl             */ 
			left = IsKeyDown(VK_LCONTROL);
			right = IsKeyDown(VK_RCONTROL);
			break;
    case 'M':
		case 'm': /* context menu key */ 
			left = right = IsKeyDown(VK_APPS);
			break;
    case 'S':
		case 's': /* shift            */ 
			left = IsKeyDown(VK_LSHIFT);
			right = IsKeyDown(VK_RSHIFT);
			break;
    case 'W':
		case 'w': /* win key          */ 
			left = IsKeyDown(VK_LWIN);
			right = IsKeyDown(VK_RWIN);
			break;
    default: return 0;
		}
		i = 3;
		switch (*src++) {
		case 'l': i=1; break;
		case 'r': i=2; break;
		}
		if ((i&1)==0) left = false;
		if ((i&2)!=0) left |= right;
		/* result */ 
		if (left) *dest = '1';
    else *dest = '0';
		return 1;

  default: return 0;
  }

  ASSERT(false);
}

void LSReplaceAtVars(char *dest,int maxchars,const char *src,
                   lsAtVarProc LookupVar, LPARAM lParam)
{
  const char *varchar,*escape;
  int nCh;

  ASSERT(READABLE(dest,maxchars*sizeof(char)));
  ASSERT(WRITEABLE(dest,maxchars*sizeof(char)));
  ASSERT(READABLE(src,sizeof(char)));
  ASSERT(EXECABLE(LookupVar));

 	/* find varchar             */ 
  while (maxchars && (varchar=strchr(src,VARCHAR)))
  {
		/* copy to this point      */ 
    nCh = varchar-src;
    if (maxchars<(nCh+2)) break; /* no room */ 
		strncpy(dest,src,nCh+1);
		maxchars-=nCh; src+=nCh; dest+=nCh;
		ASSERT(src==varchar);

    /* an atvar? */ 
    if (src[1]==ATCHAR)
    {
      /* look for the end */ 
      varchar = strchr(++src,VARCHAR);
      if (!varchar)
      {
        /* nai, copy it as literal */ 
        *dest++ = VARCHAR;
        --maxchars;
      }
      else
      {
        /* replace src to varchar with atvar value */ 
        nCh = LookupVar(dest,maxchars,++src,lParam);
        if (!nCh) nCh = LookupStdVar(dest,maxchars,src);
        if (nCh>maxchars) nCh=maxchars;
        escape = strchr(src,ESCCHAR);
        if (escape && escape<varchar)
          nCh = EscapeVar(dest,maxchars,nCh,escape+1);
        maxchars -= nCh;
        dest += nCh;
        src = varchar+1;
      }
    }
    else
    {
      /* copy varchar as literal and continue */ 
      *dest++ = *src++;
      --maxchars;
    }
    /* if it's less than zero, something went very wrong */ 
    ASSERT(maxchars>=0);
  }

  /* copy anything left over */ 
	strncpy(dest,src,maxchars);
  dest[maxchars-1] = 0;
}


